#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""JSON UTs."""

from __future__ import absolute_import, unicode_literals

from unittest import main

from b3j0f.utils.ut import UTCase

from json import dumps

from ..json import JSONSchemaBuilder, _SCHEMASBYJSONNAME
from ...base import Schema


class JSONSchemaTest(UTCase):

    def setUp(self):

        self.builder = JSONSchemaBuilder()

    def test_default(self):

        resource = '{"id": "test", "property": {}, "title": "title"}'

        schema = self.builder.build(resource)

        self.assertTrue(issubclass(schema, Schema))

    def test_elementaries(self):

        resource = {'id': 'test', 'title': 'title'}

        for rtype, stype in _SCHEMASBYJSONNAME.items():

            fresource = resource.copy()
            fresource['type'] = rtype

            sresource = dumps(fresource)

            schema = self.builder.build(sresource)

            self.assertTrue(issubclass(schema, stype))

    def _test_composite(self):

        resource = {
            'type': 'object',
            'title': 'test',
            'id': 'uuid',
            'properties': {
                '': {
                    'type': 'integer',
                    'title': 'example'
                },
                'test': {
                    'type': 'boolean'
                }
            }
        }

        json = dumps(resource)

        schemacls = self.builder.build(json)

        self.assertTrue(
            issubclass(schemacls, _SCHEMASBYJSONNAME[resource['type']])
        )
        self.assertEqual(schemacls.__name__, resource['title'])

        self.assertIsInstance(
            schemacls.test,
            _SCHEMASBYJSONNAME[resource['properties']['test']['type']]
        )
        self.assertIsInstance(
            schemacls.example,
            _SCHEMASBYJSONNAME[resource['properties']['']['type']]
        )
        self.assertFalse(hasattr(schemacls.example))

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Python language schemas utilities."""

from re import compile as re_compile

from b3j0f.utils.version import OrderedDict
from b3j0f.utils.path import lookup

from ..base import Schema
from .factory import SchemaBuilder, build
from ..elementary import ElementarySchema, ArraySchema, OneOfSchema
from ..utils import (
    updatecontent, data2schema, datatype2schemacls, RefSchema
)

from types import (
    FunctionType, MethodType, LambdaType, BuiltinFunctionType,
    BuiltinMethodType, MemberDescriptorType
)

from six import get_function_globals

from inspect import getargspec, getsourcelines, isclass, isbuiltin

from functools import wraps

__all__ = [
    'PythonSchemaBuilder', 'FunctionSchema', 'buildschema', 'ParamSchema'
]


class PythonSchemaBuilder(SchemaBuilder):
    """In charge of building python classes."""

    __name__ = 'python'

    def build(self, _resource, **kwargs):

        if not isclass(_resource):
            raise TypeError(
                'Wrong type {0}, \'type\' expected'.format(_resource)
            )

        if issubclass(_resource, Schema):
            result = _resource

        else:
            result = datatype2schemacls(_datatype=_resource, _force=False)

            if result is None:
                resname = _resource.__name__

                if 'name' not in kwargs:
                    kwargs['name'] = resname

                for attrname in dir(_resource):
                    if (
                        attrname and attrname[0] != '_' and
                        attrname not in kwargs and
                        not hasattr(Schema, attrname)
                    ):
                        attr = getattr(_resource, attrname)

                        if not isinstance(attr, MemberDescriptorType):
                            kwargs[attrname] = attr

                result = type(resname, (Schema,), kwargs)

        return result

    def getresource(self, schemacls):

        result = None

        if hasattr(schemacls, 'mro'):
            for mro in schemacls.mro():

                if issubclass(mro, Schema):
                    result = mro
                    break

        return result


def buildschema(_cls=None, **kwargs):
    """Class decorator used to build a schema from the decorate class.

    :param type _cls: class to decorate.
    :param kwargs: schema attributes to set.
    :rtype: type
    :return: schema class.
    """
    if _cls is None:
        return lambda _cls: buildschema(_cls=_cls, **kwargs)

    result = build(_cls, **kwargs)

    return result


@updatecontent
class ParamSchema(RefSchema):
    """Function parameter schema."""

    #: if true (default), update self ref when default is given.
    autotype = True
    mandatory = True  #: if true (default), parameter value is mandatory.

    def _setvalue(self, schema, value, *args, **kwargs):

        super(ParamSchema, self)._setvalue(schema, value, *args, **kwargs)

        if schema.name == 'default':

            if self.autotype and self.ref is None:
                self.ref = None if value is None else data2schema(value)

            if value is not None:
                self.mandatory = False


class FunctionSchema(ElementarySchema):
    """Function schema.

    Dedicated to describe functions, methods and lambda objects.
    """

    _PDESC = r':param (?P<ptype1>[\w_,]+) (?P<pname1>\w+):'
    _PTYPE = r':type (?P<pname2>[\w_]+):(?P<ptype2>[^\n]+)'
    _RTYPE = r':rtype:(?P<rtype>[^\n]+)'

    _REC = re_compile('{0}|{1}|{2}'.format(_PDESC, _PTYPE, _RTYPE))

    __data_types__ = [
        FunctionType, MethodType, LambdaType, BuiltinFunctionType,
        BuiltinMethodType
    ]

    params = ArraySchema(itemtype=ParamSchema())
    rtype = Schema()
    impl = ''
    impltype = ''
    safe = False

    def _validate(self, data, owner, *args, **kwargs):

        ElementarySchema._validate(self, data=data, *args, **kwargs)

        if data != self.default or data is not self.default:

            if data.__name__ != self.name:

                raise TypeError(
                    'Wrong function name {0}. {1} expected.'.format(
                        data.__name__, self.name
                    )
                )

            params, rtype = self._getparams_rtype(function=data)

            if len(params) != len(self.params):
                raise TypeError(
                    'Wrong param length: {0}. {1} expected.'.format(
                        len(params), len(self.params)
                    )
                )

            if self.rtype is not None:
                self.rtype._validate(data=rtype)

            for index, pkwargs in enumerate(params.values()):
                name = pkwargs['name']
                default = pkwargs.get('default')
                selfparam = self.params[index]

                if selfparam.name != name:
                    raise TypeError(
                        'Wrong parameter {0} at {1}. {2} expected.'.format(
                            name, index, selfparam.name
                        )
                    )

                if selfparam.default != default:
                    raise TypeError(
                        'Wrong default value {0} at {1}. Expected {2}.'.format(
                            default, index, selfparam.default
                        )
                    )

    def _setvalue(self, schema, value):

        if schema.name == 'default':
            self._setter(obj=self, value=value)

    def _setter(self, obj, value, *args, **kwargs):

        if hasattr(self, 'olddefault'):
            if self.olddefault is value:
                return

        self.olddefault = value

        ElementarySchema._setter(self, obj, value, *args, **kwargs)

        pkwargs, self.rtype = self._getparams_rtype(value)

        params = []

        selfparams = {}

        for selfparam in self.params:
            selfparams[selfparam.name] = selfparam

        index = 0

        for index, pkwarg in enumerate(pkwargs.values()):

            name = pkwarg['name']

            selfparam = None  # old self param

            if name in selfparams:
                selfparam = selfparams[name]

            if selfparam is None:
                selfparam = ParamSchema(**pkwarg)

            else:
                for key in pkwarg:
                    val = pkwarg[key]
                    if val is not None:
                        setattr(selfparam, key, val)

            params.append(selfparam)

        self.params = params

        self.impltype = 'python'

        try:
            self.impl = str(getsourcelines(value))

        except TypeError:
            self.impl = ''

    @classmethod
    def _getparams_rtype(cls, function):
        """Get function params from input function and rtype.

        :return: OrderedDict or param schema by name and rtype.
        :rtype: tuple
        """
        try:
            args, _, _, default = getargspec(function)

        except TypeError:
            args, _, _, default = (), (), (), ()

        indexlen = len(args) - (0 if default is None else len(default))

        params = OrderedDict()

        for index, arg in enumerate(args):

            pkwargs = {
                'name': arg,
                'mandatory': True
            }  # param kwargs

            if index >= indexlen:  # has default value
                value = default[index - indexlen]
                pkwargs['default'] = value
                pkwargs['ref'] = None if value is None else data2schema(value)
                pkwargs['mandatory'] = False

            params[arg] = pkwargs

        rtype = None

        # parse docstring
        if function.__doc__ is not None and not isbuiltin(function):

            scope = get_function_globals(function)

            for match in cls._REC.findall(function.__doc__):
                if rtype is None:
                    rrtype = match[4].strip() or None

                    if rrtype:

                        rtypes = rrtype.split(',')

                        schemas = []
                        for rtype_ in rtypes:

                            rtype_ = rtype_.strip()
                            islist = False

                            try:
                                lkrtype = lookup(rtype_, scope=scope)

                            except ImportError:
                                islist = True

                                try:
                                    if rtype_[-1] == 's':
                                        lkrtype = lookup(
                                            rtype_[:-1], scope=scope
                                        )

                                    elif rtype_.startswith('list of '):
                                        lkrtype = lookup(
                                            rtype_[8:], scope=scope
                                        )

                                    else:
                                        raise

                                except ImportError:
                                    msg = 'rtype {0}({1}) from {1} not found.'
                                    raise ImportError(
                                        msg.format(rtype_, rrtype, function)
                                    )

                            schemacls = datatype2schemacls(lkrtype)
                            rschema = schemacls()

                            if islist:
                                rschema = ArraySchema(itemtype=rschema)

                            schemas.append(rschema)

                        if len(rtypes) > 1:
                            rtype = OneOfSchema(schemas=schemas, nullable=True)

                        else:
                            rtype = schemas[0]

                        continue

                pname = (match[1] or match[2]).strip()

                if pname and pname in params:

                    ptype = (match[0] or match[3]).strip()

                    ptypes = ptype.split(',')

                    schemas = []

                    for ptype in ptypes:

                        ptype = ptype.strip()

                        islist = False

                        try:
                            lkptype = lookup(ptype, scope=scope)

                        except ImportError:
                            islist = True

                            try:
                                if ptype[-1] == 's':
                                    lkptype = lookup(ptype[:-1], scope=scope)

                                elif ptype.startswith('list of '):
                                    lkptype = lookup(ptype[8:], scope=scope)

                                else:
                                    raise

                            except ImportError:

                                msg = 'Error on ptype {0}("{1}") from {2}'
                                raise ImportError(
                                    msg.format(ptype, pname, function)
                                )

                        schemacls = datatype2schemacls(lkptype)
                        pschema = schemacls()

                        if islist:
                            pschema = ArraySchema(itemtype=pschema)

                        schemas.append(pschema)

                    if len(ptypes) > 1:
                        pschema = OneOfSchema(schemas=schemas, nullable=True)

                    else:
                        pschema = schemas[0]

                    params[pname]['ref'] = pschema

        return params, rtype

    def __call__(self, *args, **kwargs):

        return self.default(*args, **kwargs)

    def _getter(self, obj, *args, **kwargs):

        func = ElementarySchema._getter(self, obj, *args, **kwargs)

        @wraps(func)
        def result(*args, **kwargs):

            try:
                result = func(obj, *args, **kwargs)

            except TypeError:
                result = func(*args, **kwargs)

            return result

        result.source = func

        return result


def funcschema(default=None, *args, **kwargs):
    """Decorator to use in order to transform a function into a schema."""
    if default is None:
        return lambda default: funcschema(default=default, *args, **kwargs)

    return FunctionSchema(default=default, *args, **kwargs)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Base schema package."""

from b3j0f.utils.version import OrderedDict

from inspect import getmembers

from six import iteritems

from uuid import uuid4

__all__ = ['Schema', 'DynamicValue']


class DynamicValue(object):
    """Handle a function in order to dynamically lead a value while cleaning a
    schema.

    For example, the schema attribute ``uuid`` uses a DynamicValue in order to
    ensure default generation per instanciation.
    """

    __slots__ = ['func']

    def __init__(self, func, *args, **kwargs):
        """:param func: function to execute while cleaning a schema."""
        super(DynamicValue, self).__init__(*args, **kwargs)

        self.func = func

    def __call__(self):

        return self.func()


class Schema(property):
    """Schema description.

    A schema is identified by a string such as an universal unique identifier,
    and optionnally a name.

    Any setted value respect those conditions in this order:
    1. if the value is a lambda expression, the value equals its execution.
    2. the value is validated with this method `validate`.
    3. the value is given to a custom setter (`fget` constructor parameter) if
        given or setted to this attribute `_value`.

    Once you defined your schema inheriting from this class, your schema will
    be automatically registered in the registry and becomes accessible from the
    `b3j0f.schema.reg.getschemabyuid` function.
    """

    name = ''  #: schema name. Default is self name.
    #: schema universal unique identifier.
    uuid = DynamicValue(lambda: str(uuid4()))
    doc = ''  #: schema description.
    default = None  #: schema default value.
    required = []  #: required schema names.
    version = '1'  #: schema version.
    nullable = True  #: if True (default), value can be None.

    def __init__(
            self, fget=None, fset=None, fdel=None, doc=None, **kwargs
    ):
        """Instance attributes are setted related to arguments or inner schemas.

        :param default: default value. If lambda, called at initialization.
        """
        super(Schema, self).__init__(
            fget=self._getter, fset=self._setter, fdel=self._deleter,
            doc=doc
        )

        # set custom getter/setter/deleter
        if fget or not hasattr(self, '_fget_'):
            self._fget_ = fget

        if fset or not hasattr(self, '_fset_'):
            self._fset_ = fset

        if fdel or not hasattr(self, '_fdel_'):
            self._fdel_ = fdel

        if doc is not None:
            kwargs['doc'] = doc

        cls = type(self)

        # set inner schema values
        for name, member in getmembers(cls):

            if name[0] != '_' and name not in [
                    'fget', 'fset', 'fdel', 'setter', 'getter', 'deleter',
                    'default'
            ]:
                if name in kwargs:
                    val = kwargs[name]

                else:
                    val = member

                    if isinstance(val, DynamicValue):
                        val = val()

                    if isinstance(val, Schema):
                        val = val.default

                if isinstance(val, DynamicValue):
                    val = val()

                setattr(self, self._attrname(name=name), val)

                if member != val:
                    setattr(self, name, val)

        default = kwargs.get('default', self.default)

        self._default_ = default

        if default is not None:
            self.default = default

    def _attrname(self, name=None):
        """Get attribute name to set in order to keep the schema value.

        :param str name: attribute name. Default is this name or uuid.
        :return:
        :rtype: str
        """
        return '_{0}_'.format(name or self._name_ or self._uuid_)

    def __repr__(self):

        return '{0}({1}/{2})'.format(type(self).__name__, self.uuid, self.name)

    def __hash__(self):

        return hash(self.uuid)

    def _getter(self, obj):
        """Called when the parent element tries to get this property value.

        :param obj: parent object.
        """
        result = None

        if self._fget_ is not None:
            result = self._fget_(obj)

        if result is None:
            result = getattr(obj, self._attrname(), self._default_)

        # notify parent schema about returned value
        if isinstance(obj, Schema):
            obj._getvalue(self, result)

        return result

    def _getvalue(self, schema, value):
        """Fired when inner schema returns a value.

        :param Schema schema: inner schema.
        :param value: returned value.
        """

    def _setter(self, obj, value):
        """Called when the parent element tries to set this property value.

        :param obj: parent object.
        :param value: new value to use. If lambda, updated with the lambda
            result.
        """
        if isinstance(value, DynamicValue):  # execute lambda values.
            fvalue = value()

        else:
            fvalue = value

        self._validate(data=fvalue, owner=obj)

        if self._fset_ is not None:
            self._fset_(obj, fvalue)

        else:
            setattr(obj, self._attrname(), value)

        # notify obj about the new value.
        if isinstance(obj, Schema):
            obj._setvalue(self, fvalue)

    def _setvalue(self, schema, value):
        """Fired when inner schema change of value.

        :param Schema schema: inner schema.
        :param value: new value.
        """

    def _deleter(self, obj):
        """Called when the parent element tries to delete this property value.

        :param obj: parent object.
        """
        if self._fdel_ is not None:
            self._fdel_(obj)

        else:
            delattr(obj, self._attrname())

        # notify parent schema about value deletion.
        if isinstance(obj, Schema):
            obj._delvalue(self)

    def _delvalue(self, schema):
        """Fired when inner schema delete its value.

        :param Schema schema: inner schema.
        """

    def _validate(self, data, owner=None):
        """Validate input data in returning an empty list if true.

        :param data: data to validate with this schema.
        :param Schema owner: schema owner.
        :raises: Exception if the data is not validated.
        """
        if isinstance(data, DynamicValue):
            data = data()

        if data is None and not self.nullable:
            raise ValueError('Value can not be null')

        elif data is not None:

            isdict = isinstance(data, dict)

            for name, schema in iteritems(self.getschemas()):

                if name == 'default':
                    continue

                if name in self.required:
                    if (
                        (isdict and name not in data) or
                        (not isdict and not hasattr(data, name))
                    ):
                        part1 = (
                            'Mandatory property {0} by {1} is missing in {2}.'.
                            format(name, self, data)
                        )
                        part2 = '{0} expected.'.format(schema)
                        error = '{0} {1}'.format(part1, part2)

                        raise ValueError(error)

                elif (isdict and name in data) or hasattr(data, name):

                    value = data[name] if isdict else getattr(data, name)
                    schema._validate(data=value, owner=self)

    @classmethod
    def getschemas(cls):
        """Get inner schemas by name.

        :return: ordered dict by name.
        :rtype: OrderedDict
        """
        members = getmembers(cls, lambda member: isinstance(member, Schema))

        result = OrderedDict()

        for name, member in members:
            result[name] = member

        return result

    @classmethod
    def apply(cls, *args, **kwargs):
        """Decorator for schema application with parameters."""
        return lambda fget: cls(fget, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Python UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ...base import Schema
from ...utils import updatecontent, AnySchema, validate
from ..python import (
    FunctionSchema, buildschema, ParamSchema
)
from ...elementary import (
    StringSchema, IntegerSchema, FloatSchema, BooleanSchema, OneOfSchema,
    ArraySchema
)


class ParamSchemaTest(UTCase):

    def test_autotype(self):

        param = ParamSchema(default=1)

        self.assertIsInstance(param.ref, IntegerSchema)

    def test_notautotype(self):

        param = ParamSchema(autotype=False, default=1)

        self.assertIsNone(param.ref)

        self.assertRaises(
            TypeError, ParamSchema, ref=IntegerSchema, default=''
        )

    def test_autotype_dynamique(self):

        param = ParamSchema()

        param.default = 1

        self.assertIsInstance(param.ref, IntegerSchema)

        param.ref = None

        self.assertEqual(param.default, 1)

    def test_notautotype_dynamique(self):

        param = ParamSchema(autotype=False)

        param.default = 1

        self.assertIsNone(param.ref)

    def test_default(self):

        param = ParamSchema()

        self.assertIsNone(param.ref)

        param.default = 2
        self.assertEqual(param.default, 2)

    def test_ref(self):

        param = ParamSchema(ref=StringSchema())

        param.default = 'test'

        self.assertEqual(param.default, 'test')

        self.assertRaises(TypeError, setattr, param, 'default', 2)


class FunctionSchemaTest(UTCase):

    def test_lambda(self):

        schema = FunctionSchema(default=lambda: None)

        self.assertFalse(schema.params)

    def test_lambda_params(self):

        schema = FunctionSchema(default=lambda a, b=2: None)

        self.assertTrue(schema.params)

    def test_function(self):

        class A(object):
            pass

        def test(a, b, c, d, e=3., f=None, g=None, *args, **kwargs):
            """
            :param bool b:
            :type c: int
            :param AnySchema g: self test.
            :rtype: bool
            """

        schema = FunctionSchema(
            default=test,
            params=[
                ParamSchema(name='d', default=3, mandatory=False),
                ParamSchema(name='c', mandatory=True)
            ]
        )

        self.assertEqual(len(schema.params), 7)

        aparam = schema.params[0]
        self.assertIsNone(aparam.ref)
        self.assertTrue(aparam.mandatory)
        self.assertEqual(aparam.name, 'a')
        self.assertIsNone(aparam.default)

        bparam = schema.params[1]
        self.assertIsInstance(bparam.ref, BooleanSchema)
        self.assertFalse(bparam.mandatory)
        self.assertEqual(bparam.name, 'b')
        self.assertIs(bparam.default, False)

        cparam = schema.params[2]
        self.assertIsInstance(cparam.ref, IntegerSchema)
        self.assertTrue(cparam.mandatory)
        self.assertEqual(cparam.name, 'c')
        self.assertIs(cparam.default, 0)

        dparam = schema.params[3]
        self.assertIsInstance(dparam.ref, IntegerSchema)
        self.assertTrue(dparam.mandatory)
        self.assertEqual(dparam.name, 'd')
        self.assertIs(dparam.default, 3)

        eparam = schema.params[4]
        self.assertIsInstance(eparam.ref, FloatSchema)
        self.assertFalse(eparam.mandatory)
        self.assertEqual(eparam.name, 'e')
        self.assertIs(eparam.default, 3.)

        fparam = schema.params[5]
        self.assertIsNone(fparam.ref)
        self.assertFalse(fparam.mandatory)
        self.assertEqual(fparam.name, 'f')
        self.assertIsNone(fparam.default)

        gparam = schema.params[6]
        self.assertIsInstance(gparam.ref, AnySchema)
        self.assertFalse(gparam.mandatory)
        self.assertEqual(gparam.name, 'g')
        self.assertIsNone(gparam.default)

        self.assertIsInstance(schema.rtype, BooleanSchema)

    def test_method(self):

        @updatecontent
        class Test(Schema):

            def test(self):
                """
                :param BuildSchemaTest self:
                """

                return self

        test = Test()

        self.assertIsInstance(Test.test, FunctionSchema)
        self.assertNotIsInstance(test.test, FunctionSchema)

        res = Test.test(test)

        self.assertIs(res, test)

        res = test.test()

        self.assertIs(res, test)

        param = Test.test.params[0]

        self.assertEqual(param.name, 'self')
        self.assertIsInstance(param.ref, Schema)

    def test_or(self):

        def test_or(self):
            """
            :param float,int self:
            :rtype: float,int
            """

        schema = FunctionSchema(default=test_or)

        self.assertIsInstance(schema.params[0].ref, OneOfSchema)
        validate(schema.params[0].ref, 1.)
        validate(schema.params[0].ref, 1)

        self.assertIsInstance(schema.rtype, OneOfSchema)
        validate(schema.rtype, 1)
        validate(schema.rtype, 1.)

    def test_lists(self):

        def test_lists(self):
            """
            :param ints self:
            :rtype: ints
            """

        schema = FunctionSchema(default=test_lists)

        self.assertIsInstance(schema.params[0].ref, ArraySchema)
        validate(schema.params[0].ref, [1, 2])
        self.assertRaises(Exception, validate, schema.params[0].ref, [1., 2.])

        self.assertIsInstance(schema.rtype, ArraySchema)
        validate(schema.rtype, [1, 2])
        self.assertRaises(Exception, validate, schema.rtype, [1., 2.])

    def test_list_of(self):

        def test_list_of(self):
            """
            :param ints self:
            :rtype: ints
            """

        schema = FunctionSchema(default=test_list_of)

        self.assertIsInstance(schema.params[0].ref, ArraySchema)
        validate(schema.params[0].ref, [1, 2])
        self.assertRaises(Exception, validate, schema.params[0].ref, [1., 2.])

        self.assertIsInstance(schema.rtype, ArraySchema)
        validate(schema.rtype, [1, 2])
        self.assertRaises(Exception, validate, schema.rtype, [1., 2.])


class BuildSchemaTest(UTCase):

    def test_default(self):

        @buildschema
        class Test(object):
            pass

        test = Test()

        self.assertIsInstance(test, Schema)
        self.assertTrue(issubclass(Test, Schema))
        self.assertEqual(test.name, 'Test')

    def test_name(self):

        @buildschema(name='test')
        class Test(object):
            pass

        test = Test()
        self.assertIsInstance(test, Schema)
        self.assertTrue(issubclass(Test, Schema))
        self.assertEqual(test.name, 'test')


class TestY(object):

    def test(self):
        """
        :param TestY self:
        :rtype: TestY
        """

    @staticmethod
    def tests(a):
        """
        :param TestY a:
        :rtype: TestY
        """

    @classmethod
    def testc(cls):
        """
        :param TestY cls:
        :rtype: TestY
        """

schemacls = buildschema(TestY)

assert isinstance(schemacls.test.rtype, schemacls)
assert isinstance(schemacls.test.params[0].ref, schemacls)
assert schemacls().test.source is schemacls.test.default

assert isinstance(schemacls.tests.rtype, schemacls)
assert isinstance(schemacls.tests.params[0].ref, schemacls)
assert schemacls().tests.source is schemacls.tests.default

assert isinstance(schemacls.testc.rtype, schemacls)
assert isinstance(schemacls.testc.params[0].ref, schemacls)
assert schemacls().testc.source is schemacls.testc.default

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""json schema module."""

from __future__ import absolute_import, unicode_literals

from copy import deepcopy

from six import string_types

from ..base import Schema
from .python import FunctionSchema
from .factory import SchemaBuilder
from ..utils import updatecontent
from ..elementary import (
    ElementarySchema,
    IntegerSchema, FloatSchema, ComplexSchema, LongSchema,
    BooleanSchema,
    ArraySchema, DictSchema,
    EnumSchema,
    StringSchema,
    DateTimeSchema
)

from json import loads, load, dumps

from os.path import exists

__all__ = ['JSONSchemaBuilder']

_SCHEMASBYJSONNAME = {
    'integer': IntegerSchema,
    'number': FloatSchema,
    'long': LongSchema,
    'complex': ComplexSchema,
    'float': FloatSchema,
    'string': StringSchema,
    'array': ArraySchema,
    'dict': DictSchema,
    'boolean': BooleanSchema,
    'function': FunctionSchema,
    'datetime': DateTimeSchema,
    'enum': EnumSchema,
    'object': Schema,
    'datetime': DateTimeSchema
}

_PARAMSBYNAME = {
    'defaultValue': 'default',
    'title': 'name',
    'id': 'uuid',
    'items': 'items'
}


def json2schema(resource, name=None):

    name = resource.pop('title', name)

    uuid = resource.pop('id', None)

    stype = resource.pop('type', 'object')

    properties = resource.pop(
        'properties', resource.pop('property', {})
    )

    content = {'name': StringSchema(default=name)}

    if uuid:
        content['uuid'] = StringSchema(default=uuid)

    for name, prop in properties.items():

        if name in _PARAMSBYNAME:
            name = _PARAMSBYNAME[name]

        innerschemacls = json2schema(prop, name=name)

        content[name] = innerschemacls()

    result = type(str(name), (_SCHEMASBYJSONNAME[stype],), content)

    updatecontent(result)

    return result


class JSONSchemaBuilder(SchemaBuilder):
    """In charge of build json schemas."""

    __name__ = 'json'

    def build(self, _resource, **kwargs):

        if isinstance(_resource, string_types):

            if exists(_resource):
                fresource = load(_resource)

            else:
                fresource = loads(_resource)

        elif isinstance(_resource, dict):
            fresource = deepcopy(_resource)

        else:
            raise TypeError('Wrong type for resource {0}'.format(_resource))

        result = json2schema(fresource)

        return result

    def getresource(self, schemacls):

        def _getdict(schema):

            result = {}

            for innerschema in schema.getschemas():

                if isinstance(innerschema, ElementarySchema):
                    val = getattr(schema, innerschema.name)

                else:
                    val = _getdict(innerschema)

                result[innerschema.name] = val

            return result

        json = _getdict(schemacls)

        dump = dumps(json)

        return dump



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Utils UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from .base import Schema
from ..utils import (
    DynamicValue, data2schema, ThisSchema, validate, updatecontent,
    RegisteredSchema, dump, RefSchema, AnySchema, data2schemacls,
    datatype2schemacls
)
from ..elementary import (
    StringSchema, IntegerSchema, TypeSchema, FloatSchema, BooleanSchema
)
from ..registry import registercls, unregistercls

from six import string_types

from numbers import Number


class UpdateContentTest(UTCase):

    def _assert(self, schemacls):

        self.assertIsInstance(schemacls.a, IntegerSchema)
        self.assertIsInstance(schemacls.b, FloatSchema)
        self.assertIsInstance(schemacls.c, StringSchema)
        self.assertIsInstance(schemacls.d, TypeSchema)
        self.assertIsInstance(schemacls.e, AnySchema)

    def test_object(self):

        class TestSchema(object):

            a = 1
            b = 2.
            c = str()
            d = object
            e = None

        updatecontent(TestSchema, updateparents=False)

        self._assert(TestSchema)

    def test_schema(self):

        class TestSchema(Schema):

            a = 1
            b = 2.
            c = str()
            d = object
            e = None

        updatecontent(TestSchema, updateparents=False)

        self._assert(TestSchema)

    def test_object_decorator(self):

        @updatecontent(updateparents=False)
        class TestSchema(object):

            a = 1
            b = 2.
            c = str()
            d = object
            e = None

        self._assert(TestSchema)

    def test_schema_decorator(self):

        @updatecontent(updateparents=False)
        class TestSchema(Schema):

            a = 1
            b = 2.
            c = str()
            d = object
            e = None

        self._assert(TestSchema)


class DumpTest(UTCase):

    def test_dump(self):

        schema = Schema()

        dumped = dump(schema)

        self.assertEqual(
            dumped,
            {
                'default': schema.default,
                'name': '',
                'uuid': schema.uuid,
                'nullable': schema.nullable,
                'required': schema.required,
                'version': schema.version,
                'doc': schema.doc
            }
        )

    def test_dumped_content(self):

        class TestSchema(RegisteredSchema):

            a = Schema(default=Schema())

            b = Schema()

        schema = TestSchema()

        dumped = dump(schema)

        self.assertEqual(
            dumped,
            {
                'a': {
                    'default': None,
                    'name': '',
                    'uuid': schema.a.uuid,
                    'nullable': schema.a.nullable,
                    'required': schema.a.required,
                    'version': schema.a.version,
                    'doc': schema.a.doc
                },
                'b': None,
                'default': schema.default,
                'name': '',
                'uuid': schema.uuid,
                'nullable': schema.nullable,
                'required': schema.required,
                'version': schema.version,
                'doc': schema.doc
            }
        )


class ValidateTest(UTCase):

    def test_validate(self):

        schema = Schema()

        validate(schema, None)
        validate(schema, 1)

        schema.nullable = False
        self.assertRaises(ValueError, validate, schema, None)

        schema.nullable = True
        validate(schema, None)


class ThisSchemaTest(UTCase):

    def test_error(self):

        def definition():

            class Test(RegisteredSchema):

                test = ThisSchema(default='test', nullable=False)

                def __init__(self, *args, **kwargs):

                    super(Test, self).__init__(*args, **kwargs)

        self.assertRaises(NameError, definition)

    def test_error_deco(self):

        def definition():

            @updatecontent
            class Test(Schema):

                __update_content__ = False

                test = ThisSchema(default='test', nullable=False)

                def __init__(self, *args, **kwargs):

                    super(Test, self).__init__(*args, **kwargs)

        self.assertRaises(NameError, definition)

    def test(self):

        class Test(Schema):

            __update_content__ = False

            test = ThisSchema(test='test', nullable=False)

            def __init__(self, *args, **kwargs):

                super(Test, self).__init__(*args, **kwargs)

        self.assertIsInstance(Test.test, ThisSchema)

        updatecontent(Test)

        self.assertIsInstance(Test.test, Test)
        self.assertEqual(Test.test._test_, 'test')
        self.assertFalse(Test.test.nullable)

    def test_params(self):

        this = ThisSchema(1, 2, a=3, b=4)

        self.assertEqual(this.args, (1, 2))
        self.assertEqual(this.kwargs, {'a': 3, 'b': 4})

    def test_default(self):

        class TestSchema(RegisteredSchema):

            default = ThisSchema()

        schema = TestSchema()
        self.assertIsInstance(schema.default, TestSchema)

        schema = TestSchema(default=None)
        self.assertIsNone(schema._default_)
        self.assertIsNone(schema.default)


class Data2SchemaTest(UTCase):

    class BaseTest(Schema):

        def __init__(self, default=None, *args, **kwargs):

            super(Data2SchemaTest.BaseTest, self).__init__(*args, **kwargs)
            self.default = default

    class Test(BaseTest):
        pass

    def setUp(self):

        registercls(
            schemacls=Data2SchemaTest.BaseTest,
            data_types=[Data2SchemaTest.BaseTest]
        )

    def tearDown(self):

        unregistercls(Data2SchemaTest.BaseTest)
        unregistercls(Data2SchemaTest.Test)
        unregistercls(object)

    def test_default(self):

        self.assertIsNone(data2schema(object()))

    def test_default_force(self):

        res = data2schema(_data=map, _force=True, name='test')

        self.assertIsNotNone(res)
        self.assertEqual(res.name, 'test')

    def test_default_besteffort(self):

        self.assertIsNone(data2schema(object(), _besteffort=False))

    def test_dynamicvalue(self):

        res = data2schema(DynamicValue(lambda: ''), name='test', _force=True)

        self.assertIsNotNone(res)
        self.assertEqual(res.name, 'test')

    def test_registered(self):

        test = Data2SchemaTest.Test()
        res = data2schema(_data=test)

        self.assertEqual(res.default, test)

    def test_registered_besteffort(self):

        test = Data2SchemaTest.Test()
        res = data2schema(_data=test, _besteffort=False)

        self.assertIsNone(res)

    def test_w_attrs(self):

        class Test(object):
            pass

        test = Test()
        test.test = 1

        res = data2schema(_data=test, _force=True)

        self.assertTrue(hasattr(res, 'test'))


class RefSchemaTest(UTCase):

    def setUp(self):

        class NumberSchema(Schema):

            default = 0

            def _validate(self, data, *args, **kwargs):

                if not isinstance(data, Number):
                    raise TypeError()

        self.numberschema = NumberSchema()

        class StringSchema(Schema):

            def _validate(self, data, *args, **kwargs):

                if not isinstance(data, string_types):
                    raise TypeError()

        self.stringschema = StringSchema()

    def test_default_noref(self):

        schema = RefSchema()

        schema.default = 0

    def test_default(self):

        schema = RefSchema(ref=self.numberschema, default=1)

        self.assertEqual(schema.default, 1)

        schema.default = 0

        self.assertEqual(schema.default, 0)

        self.assertRaises(TypeError, setattr, schema, 'default', '')

    def test_owner(self):

        schema = RefSchema()

        schema._validate(0, owner=self.numberschema)

    def test_ref(self):

        schema = RefSchema(ref=self.numberschema)

        self.assertEqual(schema.default, self.numberschema.default)

        self.assertRaises(TypeError, setattr, schema, 'ref', self.stringschema)


class Data2SchemaClsTest(UTCase):

    def test_dict(self):

        data = {
            'a': 1,
            'b': True
        }

        schemacls = data2schemacls(_data=data, name='test')

        self.assertIsInstance(schemacls.a, IntegerSchema)
        self.assertEqual(schemacls.a.default, 1)
        self.assertIsInstance(schemacls.b, BooleanSchema)
        self.assertEqual(schemacls.b.default, True)
        self.assertIsInstance(schemacls.name, StringSchema)
        self.assertEqual(schemacls.name.default, 'test')

        validate(schemacls(), data)

    def test_object(self):

        class Test(object):
            a = 1
            b = True

        schemacls = data2schemacls(_data=Test, name='test')

        self.assertIsInstance(schemacls.a, IntegerSchema)
        self.assertEqual(schemacls.a.default, 1)
        self.assertIsInstance(schemacls.b, BooleanSchema)
        self.assertEqual(schemacls.b.default, True)
        self.assertIsInstance(schemacls.name, StringSchema)
        self.assertEqual(schemacls.name.default, 'test')

        validate(schemacls(), Test)
        validate(schemacls(), Test())


class DataType2Schemacls(UTCase):

    def test_namespace(self):

        class A:

            def test(self):
                pass

        schemacls = datatype2schemacls(A)

        validate(schemacls(), A())
        self.assertTrue(hasattr(schemacls, 'test'))

    def test_cls(self):

        class A(object):

            def test(self):
                pass

        schemacls = datatype2schemacls(A)

        validate(schemacls(), A())
        self.assertTrue(hasattr(schemacls, 'test'))

    def test_slots(self):

        class A(object):

            __slots__ = ['A']

            def test(self):
                pass

        schemacls = datatype2schemacls(A)

        validate(schemacls(), A())
        self.assertTrue(hasattr(schemacls, 'test'))

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------


from unittest import main

from b3j0f.utils.ut import UTCase

from ..base import Schema, DynamicValue


class DynamicValueTest(UTCase):

    def test(self):

        dvalue = DynamicValue(lambda: 'test')

        self.assertEqual('test', dvalue())


class SchemaTest(UTCase):

    def test_init(self):

        schema = Schema()

        self.assertIsNone(schema._fget_)
        self.assertIsNone(schema._fset_)
        self.assertIsNone(schema._fdel_)

    def test_uuid(self):

        schema1 = Schema()
        schema2 = Schema()

        self.assertNotEqual(schema1.uuid, schema2.uuid)

    def test_uuid_inheritance(self):

        class BaseTest(Schema):
            pass

        class Test(BaseTest):
            pass

        basetest = BaseTest()
        test = Test()

        self.assertNotEqual(basetest.uuid, test.uuid)

    def test_init_gsd(self):

        class Test(object):

            test = Schema()

        test = Test()

        res = test.test
        self.assertIsNone(res)

        schema = Schema()

        test.test = schema
        self.assertIsInstance(test.test, Schema)

        del test.test
        self.assertFalse(hasattr(test, Test.test._attrname()))

        test.test = Schema()
        self.assertIsInstance(test.test, Schema)


class DecorateTest(UTCase):

    def setUp(self):

        class TestSchema(Schema):

            test = Schema()

            def __init__(self, *args, **kwargs):

                super(TestSchema, self).__init__(*args, **kwargs)

                self.required = ['test']

        self.testschema = TestSchema

    def _assert(self, cls, processing):

        test = cls()
        self.assertNotIn('getter', processing)
        self.assertNotIn('setter', processing)
        self.assertNotIn('deleter', processing)

        res = test.test
        self.assertEqual(res, test)
        self.assertIn('getter', processing)
        self.assertNotIn('setter', processing)
        self.assertNotIn('deleter', processing)

        test.test = self.testschema()
        self.assertIsInstance(test.test, Schema)
        self.assertIn('getter', processing)
        self.assertIn('setter', processing)
        self.assertNotIn('deleter', processing)

        test.test = None
        cls.test.nullable = False
        self.assertRaises(ValueError, setattr, test, 'test', None)

        self.assertRaises(ValueError, setattr, test, 'test', 1)

        test.test = DynamicValue(lambda: self.testschema())

        self.assertRaises(ValueError, setattr, test, 'test', lambda: None)
        self.assertRaises(ValueError, setattr, test, 'test', lambda: 1)

        cls.test.nullable = True
        test.test = DynamicValue(lambda: None)

        del test.test
        self.assertFalse(hasattr(test, '_value'))
        self.assertIn('getter', processing)
        self.assertIn('setter', processing)
        self.assertIn('deleter', processing)

        test.test = self.testschema()
        self.assertIsInstance(test.test, Schema)
        self.assertIn('getter', processing)
        self.assertIn('setter', processing)
        self.assertIn('deleter', processing)

    def test_init_gsd_custom(self):

        processing = []

        class Test(object):

            @self.testschema
            def test(self):
                processing.append('getter')
                return getattr(self, '_value', self)

            @test.setter
            def test(self, value):
                processing.append('setter')
                self._value = value

            @test.deleter
            def test(self):
                processing.append('deleter')
                del self._value

        self._assert(Test, processing)

    def test_init_gsd_custom_params(self):

        processing = []

        class Test(object):

            @self.testschema.apply()
            def test(self):
                processing.append('getter')
                return getattr(self, '_value', self)

            @test.setter
            def test(self, value):
                processing.append('setter')
                self._value = value

            @test.deleter
            def test(self):
                processing.append('deleter')
                del self._value

        self._assert(Test, processing)

    def test__validate(self):

        schema = Schema()

        schema._validate(None)

        schema.nullable = False
        self.assertRaises(ValueError, schema._validate, None)

        schema.nullable = True
        schema._validate(None)

    def test_getschemas(self):

        class TestSchema(Schema):

            a = Schema()

            b = Schema()

        schemas = TestSchema.getschemas()

        self.assertEqual(len(Schema.getschemas()) + 2, len(schemas))

        schema = TestSchema()

        schemas = schema.getschemas()

        self.assertEqual(len(Schema.getschemas()) + 2, len(schemas))

    def test_notify_get(self):

        class TestSchema(Schema):

            test = Schema(name='test')
            schema = None
            value = None

            def _getvalue(self, schema, value):

                if schema.name == 'test':
                    self.schema = schema
                    self.value = value

        self.assertIsNone(TestSchema.schema)
        self.assertIsNone(TestSchema.value)

        schema = TestSchema()
        schema.test = Schema()
        schema.test

        self.assertIs(schema.schema, TestSchema.test)
        self.assertIs(schema.value, schema.test)

    def test_notify_set(self):

        class TestSchema(Schema):

            test = Schema(name='test')
            schema = None
            value = None

            def _setvalue(self, schema, value):

                if schema.name == 'test':
                    self.schema = schema
                    self.value = value

        self.assertIsNone(TestSchema.schema)
        self.assertIsNone(TestSchema.value)

        schema = TestSchema()
        schema.test = Schema()

        self.assertIs(schema.schema, TestSchema.test)
        self.assertIs(schema.value, schema.test)

    def test_notify_del(self):

        class TestSchema(Schema):

            test = Schema(name='test')
            schema = None

            def _delvalue(self, schema, *args, **kwargs):

                if schema.name == 'test':
                    self.schema = schema

        self.assertIsNone(TestSchema.schema)

        schema = TestSchema()
        del schema.test

        self.assertIs(schema.schema, TestSchema.test)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------


from unittest import main

from b3j0f.utils.ut import UTCase

from ..base import Schema
from ..registry import (
    SchemaRegistry, registercls, getbydatatype, unregistercls
)

from uuid import uuid4

from numbers import Number


class AAA(object):
    pass


class UpdateContentTest(UTCase):

    def setUp(self):

        class AAA(object):
            pass

        self.AAA = AAA

        class BBB(object):
            pass

        self.BBB = BBB

        class CCC(object):
            pass

        self.CCC = CCC

        @registercls([AAA])
        class AAASchema(Schema):

            def _validate(self, data, *args, **kwargs):

                return isinstance(data, AAA)

        self.AAASchema = AAASchema

        @registercls([BBB])
        class BBBSchema(Schema):

            def _validate(self, data, *args, **kwargs):

                return isinstance(data, BBB)

        self.BBBSchema = BBBSchema

        @registercls([CCC])
        class CCCSchema(Schema):

            def _validate(self, data, *args, **kwargs):

                return isinstance(data, CCC)

        self.CCCSchema = CCCSchema

    def tearDown(self):

        unregistercls(self.AAASchema)
        unregistercls(self.BBBSchema)
        unregistercls(self.CCCSchema)

    def test_number(self):

        schemacls = getbydatatype(self.AAA)
        self.assertIs(schemacls, self.AAASchema)

    def test_str(self):

        schemacls = getbydatatype(self.BBB)
        self.assertIs(schemacls, self.BBBSchema)

    def test_object(self):

        schemacls = getbydatatype(self.CCC)
        self.assertIs(schemacls, self.CCCSchema)


class DefaultTest(UTCase):

    def test(self):

        class TestSchema(Schema):

            default = 0

        schema = TestSchema()
        self.assertEqual(schema.default, 0)

        schema = TestSchema(default=None)
        self.assertIsNone(schema._default_)


class TestSchema(Schema):

    def __init__(
            self, name=None, uuid=None, _=None, default=None, *args, **kwargs
    ):

        super(TestSchema, self).__init__(*args, **kwargs)

        self.name = name or TestSchema.__name__
        self.uuid = str(uuid or uuid4())
        self._testschema = _ or TestSchema(_=self)
        self.default = default

    def __hash__(self):

        return hash(self.uuid)

    def getschemas(self):

        return {'one': self, 'two': self._testschema}


class SchemaRegistryTest(UTCase):

    def setUp(self):

        self.registry = SchemaRegistry()
        self.schemas = set([TestSchema() for i in range(5)])

    def test_init(self):

        schemaregistry = SchemaRegistry()
        self.assertFalse(schemaregistry._schbyname)
        self.assertFalse(schemaregistry._schbyuuid)
        self.assertFalse(schemaregistry._schbytype)

    def test_init_w_params(self):

        schemaregistry = SchemaRegistry(schbyname=2, schbyuuid=3, schbytype=4)
        self.assertEqual(schemaregistry._schbyname, 2)
        self.assertEqual(schemaregistry._schbyuuid, 3)
        self.assertEqual(schemaregistry._schbytype, 4)

    def test_register(self):

        for schema in self.schemas:
            self.registry.register(schema)

        schemas = self.registry.getbyname(TestSchema.__name__)

        self.assertEqual(schemas, self.schemas)

        for schema in self.schemas:
            uuid = schema.uuid
            _schema = self.registry.getbyuuid(uuid)
            self.assertEqual(schema, _schema)

            self.registry.unregister(uuid)
            self.assertRaises(KeyError, self.registry.getbyuuid, uuid)

    def test_registertype(self):

        class Schema(object):

            def __init__(self, default, *args, **kwargs):
                super(Schema, self).__init__(*args, **kwargs)
                self.default = default

        class IntSchema(Schema):
            pass

        class BoolSchema(Schema):
            pass

        class AAASchema(Schema):
            pass

        self.registry.registercls(schemacls=IntSchema, data_types=[int])
        self.registry.registercls(schemacls=BoolSchema, data_types=[bool])
        self.registry.registercls(schemacls=AAASchema, data_types=[Number])

        schemacls = self.registry.getbydatatype(int)
        self.assertIs(schemacls, IntSchema)

        schemacls = self.registry.getbydatatype(bool)
        self.assertIs(schemacls, BoolSchema)

        self.registry.unregistercls(schemacls=IntSchema)
        schemacls = self.registry.getbydatatype(int)
        self.assertIs(schemacls, AAASchema)

    def test_registertype_decorator(self):

        class Schema(object):

            def __init__(self, default, *args, **kwargs):
                super(Schema, self).__init__(*args, **kwargs)
                self.default = default

        @self.registry.registercls([int])
        class IntSchema(Schema):
            pass

        @self.registry.registercls([bool])
        class BoolSchema(Schema):
            pass

        @self.registry.registercls([Number])
        class AAASchema(Schema):
            pass

        schemacls = self.registry.getbydatatype(int)
        self.assertIs(schemacls, IntSchema)

        schemacls = self.registry.getbydatatype(bool)
        self.assertIs(schemacls, BoolSchema)

        self.registry.unregistercls(schemacls=IntSchema)
        schemacls = self.registry.getbydatatype(int)
        self.assertIs(schemacls, AAASchema)

if __name__ == '__main__':
    main()
